-- AngelPlayer: Handler for CA_ANGELWINGS
-- Enables use of wings for a improved vertical mobility and slowfalling.
-- Also leaves a gust of wind behind usable by allies.

-- quick constants to recognize what the ability is doing:
local ws_cst = {
	"WS_NONE",
	"WS_PREP",
	"WS_UP",
	"WS_FREEFALL",
	"WS_FLOAT"
}
for i = 1, #ws_cst
	rawset(_G, ws_cst[i], i-1)
end

-- Functions to get the wings to ... do things!

-- Wing fancy dust stuff
rawset(_G, "AP_wingDust", function(mo, s)

	local comp = mo.floorz
	if mo.eflags & MFE_VERTICALFLIP
		comp = mo.ceilingz
	end

	if abs(mo.z - comp) < mo.scale*128

		local inwater
		-- check for water:
		for fof in mo.subsector.sector.ffloors() do

			if (fof.flags & FF_EXISTS|FF_SWIMMABLE) == (FF_EXISTS|FF_SWIMMABLE)
			and fof.topheight > comp
			and fof.bottomheight < comp
				inwater = true
				break
			end
		end

		for i = 1, 16

			local a = P_SpawnMobj(mo.x, mo.y, comp, MT_THOK)
			a.scale = mo.scale
			a.state = inwater and S_SPINDUST_BUBBLE1 or S_SPINDUST1
			a.eflags = $ | (mo.eflags & MFE_VERTICALFLIP)
			P_InstaThrust(a, (360/16)*i*ANG1, mo.scale*(inwater and s/2 or s))
		end
		return comp
	end
	return nil	-- not spawned
end)

-- Wing spawn effect for wings.
local function AP_wingSpawnEffect(w, s)
	local g = P_SpawnGhostMobj(w)
	g.destscale = w.scale*4
	g.scalespeed = w.scale/12
	g.colorized = true
	if w.target and w.target.valid
		g.color = w.target.color
	end

	if s and w.target and w.target.valid
		S_StartSound(w.target, sfx_wngapr)
	end
	-- spawn sparks:

	-- first, get some "reference" point in the "middle" of the wings
	local wingdist = 64
	local x, y = w.x + FixedMul(w.scale, wingdist*cos(w.angle)), w.y + FixedMul(w.scale, wingdist*sin(w.angle))

	for i = 1, 16
		local s = P_SpawnMobj(x + P_RandomRange(-wingdist, wingdist)*w.scale, y + P_RandomRange(-wingdist, wingdist)*w.scale, w.z + P_RandomRange(-wingdist/2, wingdist/2)*w.scale, MT_THOK)
		s.state = S_WINGSPARK1
		s.scale = w.scale
		if w.target and w.target.valid
			s.color = w.target.color
		end
		w.flags = $ & ~MF_NOGRAVITY
		w.momx = P_RandomRange(-3, 3)*w.scale
		w.momy = P_RandomRange(-3, 3)*w.scale
		w.momz = P_RandomRange(1, 4)*w.scale*P_MobjFlip(w)

		if w.eflags & MFE_VERTICALFLIP
			s.eflags = $ | MFE_VERTICALFLIP
		end
	end
end

-- have the wings follow us
local function AP_updateWings(p, g)
	local mo = p.mo
	if not mo.ap_wings or not #mo.ap_wings return end	-- no wings to update, we're good to go

	local ang = p.drawangle + mo.ap_wingangle	-- arbitrary angle
	local dist = 10

	local wingremoval

	for i = 1, 2 do
		local w = mo.ap_wings[i]
		if not w or not w.valid continue end

		-- get correct xy position:
		local x = mo.x + FixedMul(mo.scale, dist*cos(ang))
		local y = mo.y + FixedMul(mo.scale, dist*sin(ang))

		-- get correct z position: use P_SpawnMobjFromMobj for a hack:
		local dum = P_SpawnMobjFromMobj(mo, 0, 0, mo.scale*mo.ap_wingz, MT_KANADEWING)
		local z = dum.z
		if dum and dum.valid
			P_RemoveMobj(dum)	-- bye
		end

		P_TeleportMove(w, x, y, z)
		w.angle = ang
		w.scale = mo.scale

		if mo.flags2 & MF2_DONTDRAW
			w.flags2 = $|MF2_DONTDRAW
		else
			w.flags2 = $ & ~MF2_DONTDRAW
		end	-- match flashtics

		if mo.eflags & MFE_VERTICALFLIP
			w.eflags = $|MFE_VERTICALFLIP
		else
			w.flags = $ & ~MFE_VERTICALFLIP
		end	-- match mobjflip

		if g
			AP_wingSpawnEffect(w, i==1)
		end

		if w.fuse
			local translevel = 10 - w.fuse
			w.frame = ($ & ~FF_TRANSMASK) | (translevel << FF_TRANSSHIFT)

			if w.fuse == 1
				wingremoval = true
			end
		end

		ang = p.drawangle - mo.ap_wingangle	-- ready angle for 2nd wing
	end

	if wingremoval	-- fuse has expired
		for i = 1, 2
			P_RemoveMobj(mo.ap_wings[i])
		end
		mo.ap_wings = nil	-- remove wings & kill the table
	end
end

rawset(_G, "AP_spawnWings", function(p, ang, z)
	local mo = p.mo
	if mo.ap_wings and #mo.ap_wings
		mo.ap_wingangle = ang or ANG1*100
		mo.ap_wingz = z or 60	-- update this nevertheless
		return
	end	-- wings already spawned, do set the angle however

	mo.ap_wings = {}
	for i = 1, 2

		local w = P_SpawnMobj(0,0,0, MT_KANADEWING)
		w.target = mo
		w.color = mo.color
		mo.ap_wings[i] = w
	end
	mo.ap_wingangle = ang or ANG1*100	-- allow setting custom angle.
	mo.ap_wingz = z or 60

	AP_updateWings(p, true)	-- directly teleport the wings to their correct position!
end)

-- sets a frame for both wings
rawset(_G, "AP_setWingFrame", function(p, f)
	local mo = p.mo
	if not mo.ap_wings or not #mo.ap_wings return end	-- no wings to update, we're good to go

	for i = 1, 2 do
		local w = mo.ap_wings[i]
		if not w or not w.valid continue end
		w.frame = f|FF_PAPERSPRITE
		if w.fuse		-- if we set the frame of the wings, undo the fuse
			w.fuse = 0
			AP_wingSpawnEffect(w, i==1)	-- as if they had just respawned, it's fancier like this!~
		end
	end
end)

-- prepare wing despawning
rawset(_G, "AP_despawnWings", function(p, force)
	local mo = p.mo
	if not mo.ap_wings or not #mo.ap_wings return end	-- no wings to update, we're good to go

	for i = 1, 2 do
		if not mo.ap_wings[i] or not mo.ap_wings[i].valid
			continue
		end
		if mo.ap_wings[i].fuse continue end	-- already applied.
		mo.ap_wings[i].fuse = force or 10
	end
end)


-- Ability trigger handler
addHook("AbilitySpecial", function(p)

	local mo = p.mo
	if p.charability ~= CA_ANGELWINGS return end
	if p.powers[pw_carry] or P_PlayerInPain(p) return end

	-- if not using the wings: prepare the ability
	if not mo.ap_wingstate
	and not (mo.ap_handstate == HS_DIVE)
	and not (mo.ap_handstate == HS_DASH)
	and mo.health
	and not p.gotflag
		AP_spawnWings(p)			-- get our wings ready! :PoppoSeesAll:

		mo.ap_wingstate = WS_PREP	-- ready the wings
		mo.ap_wingtimer = TICRATE/4

		mo.momx = $/2
		mo.momy = $/2
		if P_MobjFlip(mo)*mo.momz < 0
			mo.momz = $/2	-- divide downwards momentum
		end

		-- save some quick values:
		mo.ap_savemomz = mo.momz
		p.ap_savestats = {p.normalspeed, p.thrustfactor, p.jumpfactor}	-- AP_wresetStats
		p.pflags = $ & ~PF_STARTJUMP	-- kills the momentum cut thing when not holding jump
	end
end)

local function AP_wresetStats(p)
	if not p.ap_savestats or not #p.ap_savestats return end

	p.normalspeed = p.ap_savestats[1] or skins[p.realmo.skin].normalspeed	-- in case of 0
	p.thrustfactor = p.ap_savestats[2] or skins[p.realmo.skin].thrustfactor
	p.jumpfactor = p.ap_savestats[3] or skins[p.realmo.skin].jumpfactor
	p.ap_savestats = nil
end

-- Make sure to reset everything on mapload:
addHook("MapLoad", do
	for p in players.iterate do
		AP_wresetStats(p)
	end
end)

-- Angel Wings handler:
rawset(_G, "AP_angelWings", function(p)
	local mo = p.mo

	AP_updateWings(p)	-- Update the wing positions

	if p.charability2 == CA2_HANDSONIC	-- special case to show wings during handsonic!
		-- either while diving or while using overdrive's dash
		if (mo.ap_handstate == HS_DIVE or mo.ap_handstate == HS_DASH and mo.ap_overdrive)

			mo.ap_wingstate = WS_NONE
			AP_spawnWings(p, ANG1*155, 40)	-- will set angle & zheight if wings are already spawned
			AP_setWingFrame(p, leveltime & 1)
			AP_updateWings(p)	-- Update the wing positions

			for i = 1, 2
				local w = mo.ap_wings[i]
				if not w or not w.valid continue end	--?
				AP_advanceGhost(w)
			end
			return	-- don't let the rest of the script cancel the wings yet
		end
	end

	if mo.ap_wingstate and mo.ap_wingstate > WS_PREP
	and (P_IsObjectOnGround(mo) or not (p.pflags & PF_JUMPED))
	or p.powers[pw_carry]
		if p.ap_savestats
			AP_wresetStats(p)
		end
		mo.ap_wingstate = WS_NONE
		AP_despawnWings(p)
	end

	if not mo.ap_wingstate	-- wtf!
		if p.ap_savestats
			AP_wresetStats(p)
		end
		AP_despawnWings(p)
	end

	if mo.ap_wingstate == WS_PREP
		mo.momz = mo.ap_savemomz
		mo.state = S_PLAY_MIDFALL	-- kinda sus but we'll do with it lol
		AP_spawnWings(p)

		if not mo.ap_wingtimer		-- put here in case we somehow get set in this state without a timer set??
			AP_wresetStats(p)
			mo.ap_wingstate = WS_UP

			-- propel yourself:
			local waterfact = mo.eflags & MFE_UNDERWATER and 2 or 1
			P_SetObjectMomZ(mo, AP_ANGELWINGSTHRUST*FRACUNIT / waterfact)
			mo.state = S_PLAY_SPRING
			mo.ap_wingtimer = TICRATE/2

			local z = AP_wingDust(mo, 32)

			if z ~= nil		-- check if it spawned
				local updraft = P_SpawnMobj(mo.x, mo.y, z, MT_KANADEUPDRAFT)
				updraft.target = mo
				updraft.eflags = mo.eflags & (MFE_UNDERWATER|MFE_VERTICALFLIP)
				updraft.fuse = TICRATE*5/2
			end
		end

		if mo.ap_wingtimer < 4
			if mo.ap_wingtimer == 3
				S_StartSound(mo, sfx_wngfl)	-- flap wings!
			end
			AP_setWingFrame(p, C)
		else
			AP_setWingFrame(p, A)
		end

		mo.ap_wingtimer = $-1

	elseif mo.ap_wingstate == WS_UP

		AP_setWingFrame(p, D)
		-- visual effect for damage box when using Angel Wings
		if mo.ap_wingtimer
			AP_advanceGhost(mo)
			if mo.ap_wings and #mo.ap_wings	-- little unclean but this will do:
				for i = 1, 2 do
					if mo.ap_wings[i] and mo.ap_wings[i].valid
						AP_advanceGhost(mo.ap_wings[i])
					end
				end
			end
			mo.ap_wingtimer = $-1
		end

		if P_MobjFlip(mo)*mo.momz <= mo.scale*4
			mo.ap_wingtimer = 0	-- instantly cancel if momentum gets too low somehow
			if not (p.cmd.buttons & BT_JUMP)	-- already holding for float
				AP_despawnWings(p)
			end
			-- also enable float state:
			mo.ap_wingstate = WS_FREEFALL
		end
	elseif mo.ap_wingstate == WS_FREEFALL


		-- enable float state:
		if p.cmd.buttons & BT_JUMP	-- holding jump

			-- in overdrive: Allow more upwards jumps to be performed so long as the button wasn't HELD:
			if mo.ap_overdrive
				if (mo.ap_wings and mo.ap_wings[1] and mo.ap_wings[1].valid and mo.ap_wings[1].fuse or not mo.ap_wings or not mo.ap_wings[1] or not mo.ap_wings[1].valid)	-- hacky check but functionnal
					mo.ap_wingstate = WS_PREP
					mo.ap_wingtimer = TICRATE/4

					if P_MobjFlip(mo)*mo.momz < 0
						mo.momz = $/2	-- divide downwards momentum
					end

					-- save some quick values:
					mo.ap_savemomz = mo.momz
					p.ap_savestats = {p.normalspeed, p.thrustfactor, p.jumpfactor}	-- AP_wresetStats
					p.pflags = $ & ~PF_STARTJUMP	-- kills the momentum cut thing when not holding jump
					return
				end
			end

			if mo.state == S_PLAY_FALL
				mo.state = S_PLAY_FLOATSTART
			else
				mo.state = S_PLAY_FLOAT
			end
			mo.ap_wingstate = WS_FLOAT
			AP_spawnWings(p)
			-- momentum cut handled below
		else
			AP_despawnWings(p)
		end
	end

	-- not elseif because we want this to run the same frame
	if mo.ap_wingstate == WS_FLOAT

		AP_setWingFrame(p, leveltime & 1)

		if leveltime % 5 == 0
			AP_wingDust(mo, 20)
		end

		if mo.state ~= S_PLAY_FLOAT or mo.state ~= S_PLAY_FLOATSTART
			mo.state = S_PLAY_FLOAT
		end

		local waterfact = mo.eflags & MFE_UNDERWATER and 2 or 1

		if mo.eflags & MFE_VERTICALFLIP
			mo.momz = min($, AP_ANGELWINGSSLOWFALL*mo.scale / waterfact)
		else
			mo.momz = max($, -AP_ANGELWINGSSLOWFALL*mo.scale / waterfact)
		end

		if not (p.cmd.buttons & BT_JUMP)	-- let go off of jump
			mo.state = S_PLAY_FLOATEND
			mo.ap_wingstate = WS_FREEFALL	-- we can start floating again later
			AP_despawnWings(p)
		end
	end

	AP_updateWings(p)	-- Update the wing positions

end)

-- can damage hook: Angel Wings does damage when going up
addHook("PlayerCanDamage", function(p, m)

	local mo = p.mo
	if mo.ap_wingstate and mo.ap_wingstate == WS_UP and mo.ap_wingtimer
		return true
	end
end)

-- getting damaged/dying removes wings
local function AP_angelWingsDamage(mo)
	if not mo.player return end
	AP_despawnWings(mo.player, 2)
	mo.ap_wingstate = WS_NONE	-- cancel all wing stuff
	AP_wresetStats(mo.player)
end
addHook("MobjDamage", AP_angelWingsDamage, MT_PLAYER)
addHook("MobjDeath", AP_angelWingsDamage, MT_PLAYER)


-- Touching a spring ends the ability:
local function AP_angelWingsSpringCollide(mo, spring)
	-- yes we need to check for z height lmfao...

	if not (spring.flags & MF_SPRING) return end

	if mo.z + mo.height >= spring.z
	and mo.z <= spring.z + spring.height

		if mo and mo.ap_wingstate

			AP_despawnWings(mo.player, 2)
			mo.ap_wingstate = WS_NONE
		end
	end
end

addHook("MobjCollide", AP_angelWingsSpringCollide, MT_PLAYER)
addHook("MobjMoveCollide", AP_angelWingsSpringCollide, MT_PLAYER)